#import <Foundation/Foundation.h>
#import "MulleAutoreleasePool.h"
#include <pthread.h>
#import "NSArray+MulleApply.h"
#import "Stopwatch.h"



id   *init_storage( unsigned int n, unsigned int loops)
{
   id             *storage;
   unsigned int   i;
   unsigned int   j;
   NSObject       *p;
   NSAutoreleasePool   *pool;
   
   pool = [[NSAutoreleasePool alloc] init];
   
   storage = malloc( n * sizeof( id));
   
   fprintf( stdout, "Preparing objects...\n");
   fflush( stdout);
   
   for( i = 0; i < n; i++)
   {
      switch( rand() % 7)
      {
         case 0 :
            p = [[NSString alloc] initWithFormat:@"%d", 1];
            break;
            
         case 1 :
            p = [[NSMutableString alloc] initWithFormat:@"%d", 1];
            break;
            
         case 2 :
            p = @"Static";
            break;
            
         case 3 :
            p = [[NSArray alloc] initWithObjects:@"1", @"2", @"3", nil];
            break;
            
         case 4 :
            p = [[NSMutableArray alloc] initWithObjects:@"1", @"2", @"3", nil];
            break;
            
         case 5 :
            p = [[NSNumber alloc] initWithUnsignedInt:1];
            break;
            
         case 6 :
            p = [[NSNumber alloc] initWithDouble:1.0];
            break;
      }
      for( j = 0; j < loops; j++)
      {
         [p retain];
         [p retain];
      }
      storage[ i] = [p retain];  /* force leak */
   }
   
   [pool release];
   
   return( storage);
}


void   benchmark_apply( unsigned int n, unsigned int loops)
{ 
   NSAutoreleasePool      *pool;
   unsigned int           i;
   NSObject               *p;
   NSMutableArray         *array;
   
   pool = [[NSAutoreleasePool alloc] init];
   
   fprintf( stdout, "Preparing objects...\n");
   fflush( stdout);

   array = [NSMutableArray arrayWithCapacity:n];
   
   for( i = 0; i < n; i++)
   {
      switch( rand() % 7)
      {
         case 0 :
            p = [[NSString alloc] initWithFormat:@"%d", 1];
            break;
            
         case 1 :
            p = [[NSMutableString alloc] initWithFormat:@"%d", 1];
            break;
            
         case 2 :
            p = @"Static";
            break;
            
         case 3 :
            p = [[NSArray alloc] initWithObjects:@"1", @"2", @"3", nil];
            break;
            
         case 4 :
            p = [[NSMutableArray alloc] initWithObjects:@"1", @"2", @"3", nil];
            break;
            
         case 5 :
            p = [[NSNumber alloc] initWithUnsignedInt:1];
            break;
            
         case 6 :
            p = [[NSNumber alloc] initWithDouble:1.0];
            break;
      }
      [array addObject:p];
      [p release];
   }
   
//   NSLog( @"Using Foundation makePerformObjects:@selector( self) %u times on %u objects", loops, n);
   for( i = 0; i < loops; i++)
      [array makeObjectsPerformSelector:@selector( self)];
//   NSLog( @"Using mulleMakePerformObjects:@selector( self) %u times on %u objects", loops, n);
   for( i = 0; i < loops; i++)
      [array mulleMakeObjectsPerformSelector:@selector( self)];
//   NSLog(@"Done");

   [pool release];
}



void   benchmark( unsigned int n, unsigned int loops)
{ 
   NSAutoreleasePool      *pool1;
   MulleAutoreleasePool   *pool2;
   unsigned int           i;
   unsigned int           j;
   id                     *storage;
   Stopwatch              *watch;
   
   watch = [Stopwatch new];
   
//   [MulleAutoreleasePool poseAsClass:[NSAutoreleasePool class]];
   
   pool1 = [[NSAutoreleasePool alloc] init];
   pool2 = [[MulleAutoreleasePool alloc] init];

   storage = init_storage( n, loops);

   [watch startAndLog:@"Adding objects to NSAutoreleasePool"]; 
   for( j = 0; j < loops; j++)
      for( i = 0; i < n; i++)
         [pool1 addObject:storage[ i]];
   [watch stopAndLog]; 

   [watch startAndLog:@"Killing NSAutoreleasePool"]; 
   [pool1 release];
   [watch stopAndLog]; 
   
   sleep( 5);
   
   [watch startAndLog:@"Adding objects to MulleAutoreleasePool"]; 
   for( j = 0; j < loops; j++)
      for( i = 0; i < n; i++)
         [pool2 addObject:storage[ i]];
   [watch stopAndLog]; 

   [watch startAndLog:@"Killing MulleAutoreleasePool"]; 
   [pool2 release];
   [watch stopAndLog]; 
   
   [watch release];
   
//   NSLog(@"Done");
}


void   benchmark2( unsigned int n, unsigned int loops)
{ 
   NSAutoreleasePool      *pool1;
   MulleAutoreleasePool   *pool2;
   id                     *storage;
   unsigned int           i;
   unsigned int           j;
   Stopwatch              *watch;
   
   watch = [Stopwatch new];
   
   //   [MulleAutoreleasePool poseAsClass:[NSAutoreleasePool class]];
   
   [[NSAutoreleasePool alloc] init];
   
   storage = init_storage( n, loops);
   
   [watch startAndLog:@"Adding and killing in NSAutoreleasePool"]; 
   for( j = 0; j < loops; j++)
   {
      pool1 = [[NSAutoreleasePool alloc] init];
      for( i = 0; i < n; i++)
         [pool1 addObject:storage[ i]];
      [pool1 release];
   }
   [watch stopAndLog]; 

   sleep( 10);

   [watch startAndLog:@"Adding and killing objects in MulleAutoreleasePool"]; 
   for( j = 0; j < loops; j++)
   {
      pool2 = [[MulleAutoreleasePool alloc] init];
      for( i = 0; i < n; i++)
         [pool2 addObject:storage[ i]];
      [pool2 release];
   }
   [watch stopAndLog]; 
   
   [watch release];
}



@interface StoogeObject : NSObject
{
   unsigned int  no_;
}

+ (void) reset;
+ (void) multiThreadTest:(NSDate *) death;
+ (void) printStatistics;

@end


@implementation StoogeObject

static unsigned long  allocated_stooges;
static unsigned long  deallocated_stooges;

+ (void) reset
{
   allocated_stooges   = 0;
   deallocated_stooges = 0;
}


+ (void) printStatistics
{
   fprintf( stderr, "%lu allocated / %lu deallocated\n", allocated_stooges, deallocated_stooges);
}


- init
{
   [super init];
   
   no_ = ++allocated_stooges;
   
   return( self);
}


- (NSString *) description
{
   return( [NSString stringWithFormat:@"<%@ %p>", isa, self]);
}


- (void) dealloc
{
   ++deallocated_stooges;
   [super dealloc];
}


+ (void) multiThreadTest:(NSDate *) death
{ 
   NSAutoreleasePool   *pools[ 4];
   int                 i;
   BOOL                flag;
   NSCalendarDate      *now;
   
   fprintf( stderr, "%p thread runs\n", pthread_self());
   do
   {
      for( i = 0; i < 4; i++)
      {
         pools[ i] = [[NSAutoreleasePool alloc] init];
         [[StoogeObject new] autorelease];
      }
      
      for( i = 3; i >= 0; i--)
         [pools[ i] release];
      
      pools[ 0] = [[NSAutoreleasePool alloc] init];
      now  = [NSCalendarDate date];
      flag = [now compare:death] == NSOrderedAscending;
      [pools[ 0]  release];
   }
   while( flag);
   fprintf( stderr, "%p thread dies\n", pthread_self());
}

@end


void   single_thread_test( unsigned int stepdown)
{ 
   NSAutoreleasePool   *pools[ 128];
   int                 i;

   [StoogeObject reset];
   for( i = 0; i < 127; i++)
   {
      pools[ i] = [[NSAutoreleasePool alloc] init];
      NSLog( @"Added %@ to nested pool %@", [[StoogeObject new] autorelease], [pools[ i] description]);
   }
   
   for( i = 127; i >= 0; i -= stepdown)
   {
      [pools[ i] release];
   }

   if( i + stepdown != 0)
      [pools[ 0] release];
   [StoogeObject printStatistics];
}


/* fork randomly threads. Let them create and uncreate autorelease pools
 * add a few random objects in there.
 */


void  multi_thread_test( int threads)
{
   NSAutoreleasePool   *pool;
   NSCalendarDate      *death;
   NSCalendarDate      *now;
   unsigned int        i;
   BOOL                flag;
   
   // multiple thread test   
   pool = [[NSAutoreleasePool alloc] init];
   
   death = [[NSCalendarDate date] dateByAddingYears:0 months:0 days:0 hours:0 minutes:0 seconds:20];
   for( i = 0; i < threads; i++)
      [NSThread detachNewThreadSelector:@selector( multiThreadTest:)
                               toTarget:[StoogeObject class]
                             withObject:death];
   
   do
   {
      sleep( 1);
      
      now  = [[NSCalendarDate alloc] init];
      flag = [now compare:death] == NSOrderedAscending;
      [now release];
   }
   while( flag);
   
   [pool release];
}


int main (int argc, const char * argv[]) 
{
   
   if( argc >= 2)
   {
      fprintf( stderr, "Sleeping for 9 seconds\n");
      sleep( 9);
      fprintf( stderr, "Sleeping for 1 second\n");
      sleep( 1);
   }
//   benchmark_apply( 10000, 10000);
//    benchmark( 1000, 50000);  /* this pains Foundation */
   benchmark( 5000, 2000);  
   benchmark2( 5000, 2000);  
   
// single thread test   
//  [MulleAutoreleasePool poseAsClass:[NSAutoreleasePool class]];
//   single_thread_test( 1);
//   single_thread_test( 3);

//   multi_thread_test( 10);
   
   if( argc >= 2)
   {
      fprintf( stderr, "Sleeping for 5 seconds\n");
      sleep( 5);
   }
   return 0;
}


